/*
 * ufdbGuard is copyrighted (C) 2005,2006 by URLfilterDB with all rights reserved.
 */

#include "ufdb.h"
#include "ufdblib.h"

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <pthread.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/stat.h>


int UFDBglobalDebug = 0;

static long          num_cpus;
static unsigned long cpu_mask = 0UL;

pthread_mutex_t ufdb_malloc_mutex = PTHREAD_MUTEX_INITIALIZER;



void * ufdbMalloc( size_t elsize )
{
  void * p;

  pthread_mutex_lock( &ufdb_malloc_mutex );
  p = (void *) malloc(elsize);
  pthread_mutex_unlock( &ufdb_malloc_mutex );

#if 0
  fprintf( stderr, "          ufdbMalloc( %12d )\n", elsize ); 
#endif

  return p;
}


void * ufdbCalloc( size_t n, size_t num )
{
  void * p;

  pthread_mutex_lock( &ufdb_malloc_mutex );
  p = (void *) calloc( n, num );
  pthread_mutex_unlock( &ufdb_malloc_mutex );

  return p;
}



void * ufdbRealloc( void * ptr, size_t elsize )
{
  void * p;

  pthread_mutex_lock( &ufdb_malloc_mutex );
  p = (void *) realloc( ptr, elsize );
  pthread_mutex_unlock( &ufdb_malloc_mutex );

  return p;
}


void ufdbFree( void * ptr )
{
   if (ptr != NULL)
   {
      pthread_mutex_lock( &ufdb_malloc_mutex );
      free( ptr );
      pthread_mutex_unlock( &ufdb_malloc_mutex );
   }
}


char * ufdbStrdup( const char * s )
{
   int size = strlen( s ) + 1;
   return strcpy( ufdbMalloc(size), s );
}


/* In a multithreaded application, regcomp and regfree must be protected
 * by a malloc mutex since they call malloc and free !
 */
int UFDBregcomp( regex_t * preg, const char * regex, int cflags )
{
   extern pthread_mutex_t ufdb_malloc_mutex;
   int retval;

   pthread_mutex_lock( &ufdb_malloc_mutex );
   retval = regcomp( preg, regex, cflags );
   pthread_mutex_unlock( &ufdb_malloc_mutex );

   return retval;
}


void UFDBregfree( regex_t * preg )
{
   pthread_mutex_lock( &ufdb_malloc_mutex );
   regfree( preg );
   pthread_mutex_unlock( &ufdb_malloc_mutex );
}


/*
 *  initialise a DB (open and/or create a .ufdb file) 
 */
int UFDBloadDatabase( 
   struct UFDBmemTable * mtable, 
   char *                file    )
{
   int                   n;
   int                   in;
   struct stat           fileStat;
   char                  f[1024];

   if (file == NULL)
   {
      return UFDB_API_ERR_NULL;
   }

#if 0
   fprintf( stderr, "   UFDBloadDatabase( %s )\n", file );
#endif

   strcpy( f, file );
   in = open( f, O_RDONLY );
   if (in < 0)
   {
      strcat( f, UFDBfileSuffix );
      in = open( f, O_RDONLY );
   }

   if (in < 0)
   {
      return UFDB_API_ERR_NOFILE;
   }

   if (fstat(in,&fileStat) < 0)
   {
      close( in );
      return UFDB_API_ERR_NOFILE;
   }

   mtable->table.tag = (unsigned char *) "UNIVERSE";
   mtable->table.nNextLevels = 0;
   mtable->table.nextLevel = NULL;
   mtable->mem = ufdbMalloc( fileStat.st_size + 1 );
   mtable->tableSize = fileStat.st_size - sizeof(struct UFDBfileHeader);
   n = read( in, mtable->mem, fileStat.st_size );
   close( in );
   if (n != fileStat.st_size)
   {
      return UFDB_API_ERR_READ;
   }

   UFDBparseTableHeader( mtable );
   UFDBparseTable( mtable );

   return UFDB_API_OK;
}



struct ufdbRegExp * ufdbNewPatternBuffer( 
  char * pattern, 
  int    flags )
{
  struct ufdbRegExp * re;

  re = (struct ufdbRegExp *) ufdbMalloc( sizeof(struct ufdbRegExp) );

  re->pattern = (char *) ufdbMalloc( strlen(pattern) + 1 );
  strcpy( re->pattern, pattern );
  re->substitute = NULL;
  re->compiled = (regex_t *) ufdbCalloc( 1, sizeof(regex_t) );
  re->error = UFDBregcomp( re->compiled, pattern, flags );
  re->flags = flags;
  re->global = 0;
  re->httpcode = NULL;
  re->next = NULL;

  return re;
}


/*
 * initialize an expression list (read them from a file and do the regexp compilation)
 */
int UFDBloadExpressions( 
   struct ufdbRegExp ** exprList,
   char *               file  )
{
   FILE *               fin;
   char *               eoln;
   struct ufdbRegExp *  re;
   struct ufdbRegExp *  last;
   int                  retCode;
   char                 line[1024];

   if (exprList == NULL)
      return UFDB_API_ERR_NULL;
   *exprList = NULL;

   if (file == NULL)
      return UFDB_API_ERR_NULL;

   fin = fopen( file, "r" );
   if (fin == NULL)
      return UFDB_API_ERR_NOFILE;

   retCode = UFDB_API_OK;
   last = NULL;
   re = NULL;

   while (fgets( line, sizeof(line), fin ) != NULL)
   {
      if (line[0] == '#')         /* skip comments */
         continue;

      eoln = strchr( line, '\n' );
      if (eoln == NULL  ||  eoln == &line[0])
         continue;	/* skip empty lines and lines without a newline */
      else
      {
         if (*(eoln-1) == '\r')
	    eoln--;
      }
      *eoln = '\0';	/* get rid of the newline */

      re = ufdbNewPatternBuffer( line, REG_EXTENDED | REG_NOSUB | REG_ICASE );
      if (re->error)
      {
#if 0
	 printf( "error compiling %s: \"%s\"\n", file, line );
#endif
         retCode = UFDB_API_ERR_EXPR;
      }

      re->next = last;
      last = re;
   }

   *exprList = re;
   fclose( fin );

   return retCode;
}


/*
 * match a URL with a compiled RE.
 * return 0 if no match, 1 if there is a match.
 */
int ufdbRegExpMatch(   
   struct ufdbRegExp * regexp, 
   char *              str )
{
   struct ufdbRegExp * rp;
   int                 error;

   for (rp = regexp; rp != NULL; rp = rp->next)
   {
#if 0
      fprintf( stderr, "ufdbRegExpMatch %s %s\n", str, rp->pattern );
#endif
      if (rp->error)
         continue;
      error = regexec( rp->compiled, str, 0, 0, 0 );
      if (error == 0) 	/* match */
         return 1;
      if (error != REG_NOMATCH) 
         return UFDB_API_ERR_EXPR;
   }

   return 0;
}


void ufdbFreeRegExprList( struct ufdbRegExp * re )
{
   if (re == NULL)
      return;

   ufdbFreeRegExprList( re->next );

   ufdbFree( re->pattern );
   UFDBregfree( re->compiled );
   ufdbFree( re->compiled );
   ufdbFree( re->substitute );
   ufdbFree( re );
}


void ufdbResetCPUs( void )
{
   cpu_mask = 0UL;
}


/*
 * Linux support binding threads to a set of CPUs which prevents cache contention.
 * Freebsd has no support, but the 5.x series is recommended for best multithreaded performance.
 * On Solaris it is recommended to start ufdbguardd in a processor set.
 */
int ufdbSetThreadCPUaffinity( void )
{
#if defined(__linux__) && defined(__NR_sched_setaffinity)
   int retval;

   if (cpu_mask != 0UL)
   {
      retval = syscall( __NR_sched_setaffinity, 0, 4, &cpu_mask );
      if (retval < 0)
         return UFDB_API_ERR_ERRNO;
   }
#endif
   return UFDB_API_OK;
}


/*
 * Bind my processes and threads to a set of CPUs.
 * This increases the cache efficiency for all programs.
 */
int ufdbSetCPU( 
   char * CPUspec )		/* comma-separated CPU numbers (starting with CPU 0) */
{
   int    cpu;

#if defined(_SC_NPROCESSORS_ONLN)
   num_cpus = sysconf( _SC_NPROCESSORS_ONLN );
   /* printf( "this system has %ld CPUs\n", num_cpus ); */

#elif defined(__NR_sched_getaffinity)
   /* sched_setaffinity() is buggy on linux 2.4.x so we use syscall() instead */
   cpu = syscall( __NR_sched_getaffinity, getpid(), 4, &cpu_mask );
   /* printf( "sched_getaffinity returned %d %08lx\n", cpu, cpu_mask ); */
   if (cpu >= 0)
   {
      num_cpus = 0;
      for (cpu = 0; cpu < 32; cpu++)
	 if (cpu_mask & (1 << cpu))
	    num_cpus++;
      /* printf( "   found %d CPUs in the cpu mask\n", num_cpus ); */
   }
   else
#endif
      num_cpus = 32;

   cpu_mask = 0;
   while (*CPUspec)
   {
      if (sscanf(CPUspec,"%d",&cpu) == 1)
      {
	 if (cpu >= 0 && cpu < num_cpus)
	 {
	    cpu_mask |= (1 << cpu);
	 }
	 else
	    return UFDB_API_ERR_RANGE;
      }
      else
         return UFDB_API_ERR_RANGE;

      /* find the next CPU number */
      while (isdigit( (int) *CPUspec))
         CPUspec++;
      while (*CPUspec == ' '  ||  *CPUspec == ',')
         CPUspec++;
   }

   return UFDB_API_OK;
}

